""" Expression generator for frequently needed terms

    This module contains expression generators frequently needed
    while they certainly could be created by hand, they come up often
    enough, to generate them.

    So to create a parameter for the sum of squares of the rates in your
    model you'd use:

    >>> add_parameter_for('sum_squares', 'rates', 'sum_of_squares')

    Or for the minimal flux:

    >>> add_parameter_for('min_flux', 'fluxes', 'min')


    objects currently supported, are fluxes and concentrations. The
    operations are: 'min', 'max', 'sum', 'sum_of_squars' and 'sum_of_abs'

"""

import COPASI
from . import model_io

class OBJ:
    RATES_OF_CHANGE = 'rates'
    FLUXES = 'fluxes'
    CONCENTRATIONS = 'concentrations'

class OPP:
    MIN = 'min'
    MAX = 'max'
    SUM = 'sum'
    SUM_OF_SQUARES = 'sum_of_squares'
    SUM_OF_ABS = 'sum_of_abs'

def _get_objects(model, objects):
    """Returns a list of object reference cns chosen

    :param objects: one of the elements of the `OBJ` class.
    :param model: the model to get the references from
    :type model: COPASI.CDataModel

    """
    if type(objects) is list:
        return objects

    object_cns = []
    if objects == OBJ.FLUXES:
        m = model.getModel()
        assert (isinstance(m, COPASI.CModel))
        for r in m.getReactions():
            assert (isinstance(r, COPASI.CReaction))
            object_cns.append(r.getFluxReference().getCN())
        return object_cns

    if objects == OBJ.RATES_OF_CHANGE:
        m = model.getModel()
        assert (isinstance(m, COPASI.CModel))
        for obj in m.getMetabolites():
            assert (isinstance(obj, COPASI.CMetab))
            object_cns.append(obj.getConcentrationRateReference().getCN())
        return object_cns

    if objects == OBJ.CONCENTRATIONS:
        m = model.getModel()
        assert (isinstance(m, COPASI.CModel))
        for obj in m.getMetabolites():
            assert (isinstance(obj, COPASI.CMetab))
            object_cns.append(obj.getConcentrationReference().getCN())
        return object_cns

    return object_cns

def _nop(objects):
    return ''

def _min(objects):
    num = len(objects)
    if num == 0:
        return '0'
    if num == 1:
        return '<' + str(objects[0]) + '>'
    result = ''
    closing = 0
    for i in range(num - 1):
        result += 'min(<' + objects[i] + '>, '
        closing += 1
    result += '<' + objects[-1] + '>'
    for _ in range(closing):
        result += ')'
    return result

def _max(objects):
    num = len(objects)
    if num == 0:
        return '0'
    if num == 1:
        return '<' + str(objects[0]) + '>'
    result = ''
    closing = 0
    for i in range(num-1):
        result += 'max(<' + objects[i] + '>, '
        closing += 1
    result += '<' + objects[-1] + '>'
    for _ in range(closing):
        result += ')'
    return result

def _sum(objects):
    if not objects:
        return '0'
    return ' + '.join(['<'+str(o)+'>' for o in objects])

def _sum_of_squares(objects):
    if not objects:
        return '0'
    return ' + '.join(['<'+str(o)+'>^2' for o in objects])

def _sum_abs(objects):
    if not objects:
        return '0'
    return ' + '.join(['abs(<'+str(o)+'>)' for o in objects])

def _get_operator(operator):
    operations = {
        OPP.MAX: _max,
        OPP.MIN: _min,
        OPP.SUM_OF_ABS: _sum_abs,
        OPP.SUM: _sum,
        OPP.SUM_OF_SQUARES: _sum_of_squares
    }
    return operations.get(operator, _nop)

def _generate_expression(objects, operator, model):
    assert (isinstance(model, COPASI.CDataModel))
    object_references = _get_objects(model, objects)
    op = _get_operator(operator)
    return op(object_references)

def add_parameter_for(name, objects, operator, **kwargs):
    """ Creates a new global model parameter with the given name

    :param name: the name of the new parameter
    :param objects: the objects to choose: 'rates', 'fluxes', 'concentrations'
    :param operator: the operation for the expression: 'sum_of_squares'
    :param kwargs: optional parameters, recognized are:

        * | `model`: to specify the data model to be used (if not specified
          | the one from :func:`.get_current_model` will be taken)

    :return: the newly created parameter
    """
    model = model_io.get_model_from_dict_or_default(kwargs)
    m = model.getModel()
    assert (isinstance(m, COPASI.CModel))
    mv = m.createModelValue(name)
    if not mv:
        raise ValueError('A global parameter named ' + name + ' already exists')

    assert (isinstance(mv, COPASI.CModelValue))
    mv.setStatus(COPASI.CModelEntity.Status_ASSIGNMENT)
    mv.setExpression(_generate_expression(objects, operator, model))
    return mv



import unittest
import basico
import os

THIS_DIR = os.path.dirname(__file__)


class TestTimeCourse(unittest.TestCase):
    def setUp(self):
        basico.load_model(os.path.join(THIS_DIR, 'tellurium_sim.sbml'))

    def test_custom_output(self):
        tc = basico.run_time_course_with_output(['Time', '[S1]', '[S2]', 'Values[k1]'], 0, 5, 5)
        self.assertIsNotNone(tc)
        self.assertListEqual(tc.columns.to_list(), ['Time', '[S1]', '[S2]', 'Values[k1]'])

    def test_stochastic(self):
        basico.new_model(name='Simple Model')
        basico.add_reaction('R1', 'A ->')
        basico.add_reaction('R2', 'B -> B + C')
        basico.set_species('A', initial_particle_number=50)
        basico.set_species('B', initial_particle_number=10)
        basico.set_species('C', initial_particle_number=0)
        basico.set_reaction_parameters('(R1).k1', value=1)
        basico.set_reaction_parameters('(R2).k1', value=1)
        basico.add_event('E1', 'A <= 10', [('B', '50')])
        result = basico.run_time_course(duration=10, method='directMethod', use_numbers=True)
        self.assertIsNotNone(result)

        # test that we can set hybrid methods
        result = basico.run_time_course(duration=1, method='hybrid', use_numbers=True)
        self.assertIsNotNone(result)
        settings = basico.get_task_settings(basico.T.TIME_COURSE)
        self.assertEqual(settings['method']['name'], 'Hybrid (Runge-Kutta)')

        # test that we cannot run the time course with too many particles
        # and that the error message is correct:
        basico.set_species('A', initial_particle_number=1e+24)
        with self.assertLogs('basico', level='ERROR') as cm:
            basico.run_time_course(duration=10, method='directMethod', use_numbers=True)
            self.assertTrue(len(cm.output) > 0)
            error_message = cm.output[0]
            self.assertTrue('At least one particle number in the initial state is too big' in error_message)

    def test_stoch2(self):
        basico.new_model(name='Simple Model', substance_unit='1')
        basico.add_reaction('R1', 'A ->')
        basico.set_species('A', initial_particle_number=50)
        basico.set_reaction_parameters('(R1).k1', value=1)
        basico.add_event('E1', 'A.ParticleNumber <= 1', [('A.ParticleNumber', '50')])

        result = basico.run_time_course(duration=20, method='directMethod', use_numbers=True)
        self.assertIsNotNone(result)

    def test_general_run(self):
        dm = basico.load_example('bruss')
        plots = []
        basico.run_task(basico.T.TIME_COURSE, model=dm, plots=plots)
        self.assertEqual(len(plots), 2)
        basico.remove_datamodel(dm)

    def test_general_run_sedml(self):
        dm = basico.load_model(os.path.join(THIS_DIR, 'sedml.xml'))
        plots = []
        basico.run_scheduled_tasks(include_plots=False, include_general_plots=True, plots=plots, model=dm)
        self.assertEqual(len(plots), 1)

    def test_datahandler_for_fluxes_with_parenthesis(self):
        basico.new_model(name='Simple Model')
        basico.add_reaction('R1 (step 1)', 'A ->')
        basico.set_species('A', initial_concentration=10)
        
        result = basico.run_time_course_with_output(['Time', 
                                                     '(R1 (step 1)).Flux',
                                                     '(R1 (step 1)).k1'], 0, 5, 5)    
        
        self.assertIsNotNone(result)
        self.assertEqual(result.shape, (6, 3))


if __name__ == '__main__':
    unittest.main()
